((c-mode . ((c-file-style . "GNU")
            (indent-tabs-mode . nil)
            (c-basic-offset . 8)
            (tab-width . 8)
            (show-trailing-whitespace . t)))
(c++-mode . ((c-file-style . "GNU")
             (indent-tabs-mode . nil)
             (c-basic-offset . 8)
             (tab-width . 8)
             (show-trailing-whitespace . t)))
(vala-mode . ((c-file-style . "GNU")
              (indent-tabs-mode . nil)
              (c-basic-offset . 2)
              (tab-width . 2)
              (show-trailing-whitespace . t))))

(defun my-c-setup ()
   (c-set-offset 'innamespace [0]))
(add-hook 'c++-mode-hook 'my-c-setup)
